//This file is meant to be left as is, except for the settings

#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ESP32Servo.h>  // older version <ESP32_Servo.h>, newer version same library <ESP32Servo.h>
#include <WString.h>
#include "car.h"
#include "webPage.h"
#include "functions.h"

namespace car{
  ///////////////Settings//////////////////
  
  const char* ssid_Host     = "car";
  const char* password_Host = "wroem";
  bool passwordProtect = false;

  const char* ssid_Client     = "Buuuuuuuuuuuuuuuuuuuus";
  const char* password_Client = "langebus";
  bool connectToWiFi = false;  //The program becomes sluggish when this setting is set to true and there is no WiFi network available
  
  //                  L   L   R   R
  int motorPins[4] = {18, 5, 13, 15};


  ////////////////Defines//////////////
  bool logData = true;
  
  int throttle = 0;
  int rot     = 0;
  
  AsyncWebServer server(80);
  AsyncWebSocket ws("/ws");

  TaskHandle_t carTask;


  ///////////////Functions/////////////
  void initCar(){    
    //init the motors by attaching the pins
    initMotor();

    //setup serial
    Serial.begin(115200);
    Serial.println();

    //start wifi network host
    if(passwordProtect){
      WiFi.softAP(ssid_Host, password_Host);
    }else{
      WiFi.softAP(ssid_Host);
    }

    //connect to wifi network as client
    if(connectToWiFi) WiFi.begin(ssid_Client, password_Client);

    //print the IP´s of the device
    IPAddress HostIP = WiFi.softAPIP();
    Serial.print("IP on network " + String(ssid_Host) + ": '");
    Serial.print(HostIP);
    Serial.println("'");

    IPAddress ClientIP = WiFi.localIP();
    Serial.print("IP on network " + String(ssid_Client) + ": '");
    Serial.print(ClientIP);
    Serial.println("'");

    //configure the webserver (websockets and webpages)
    ws.onEvent(onWebsocket);
    server.addHandler(&ws);
    
    server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
      request->send(200, "text/html", webPage::pageC);
    });

    //start the webserver
    server.begin();
    
    //run setup function from functions.cpp
    setupFunction();

    // run loop function from functions.cpp
    while(true){
      loopFunction();
    }
  }

  //websocket handle function
  void onWebsocket(AsyncWebSocket * server, AsyncWebSocketClient * client, AwsEventType type, void * arg, uint8_t *data, size_t len){
    AwsFrameInfo * info = (AwsFrameInfo*)arg;

    if(type == WS_EVT_CONNECT){
      Serial.println("Client '" + String(client->id()) + "' has connected");
    } else if(type == WS_EVT_PONG){
      Serial.println("Client '" + String(client->id()) + "' has answered ping request");
    } else if(type == WS_EVT_DISCONNECT){
      Serial.println("Client '" + String(client->id()) + "' has been disconnected");
    } else if(type == WS_EVT_ERROR){
      Serial.println("Client '" + String(client->id()) + "': an error occured");
    } else if(type == WS_EVT_DATA){
      if(info->opcode == WS_TEXT){
        //text was received, this is done by the javascript code in our webpage

        //find the command sent by the js code
        int key = (char)data[0];
        
        //build the value string out of received chars
        String value;
        for(size_t i=1; i < info->len; i++){
          value += (char)data[i];
        }

        //run the code which handles the commands sent by the js code
        switch(key){
          case 't':
            throttle = value.toInt();
            break;
          case 'r':
            rot = value.toInt();
            break;
          case 'b':
            if(value.toInt() == 1){
              button1();
            }else if(value.toInt() == 2){
              button2();
            }else if(value.toInt() == 3){
              button3();
            }else if(value.toInt() == 4){
              button4();
            }
            break;
        }
        movement(throttle, rot);
      }else{
        //binary message was received, do nothing, our js code shouldn´t send binary messages
      }
    }
  }

  //init the motors by attaching the correct pins to the pwm device (internal in esp)
  void initMotor(){
    for(int i = 1; i <= 4; i++){
      ledcSetup(i+4, 5000, 8);  //use channel 5, 6, 7, 8
      ledcAttachPin(motorPins[i - 1], i+4);
    }
  }

  //convert integer into motor compatable value (by setting the min and max values)
  int toMotorSpeed(int x){
    if(x < -255) x = -255;
    if(x > 255) x = 255;
    return x;
  }

  //return the Throttle defined by webpage
  int getThrottle(){
    return throttle;
  }

  //return the rotation defined by webpage
  int getRotation(){
    return rot;
  }

  //write motor Throttle value to motor
  void writeMotorL(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(1+4, motorThrottle);
      ledcWrite(2+4, 0);
    }else{
      ledcWrite(1+4, 0);
      ledcWrite(2+4, -motorThrottle); // motorThrottle is negative, -motorThrottle is positive
    }
  }

  //write motor Throttle value to motor
  void writeMotorR(int motorThrottle){
    //make sure Throttle value is compatable with motor
    motorThrottle = toMotorSpeed(motorThrottle);

    //find movement rotation and write the speed to the correct motor pin
    if(motorThrottle > 0){
      ledcWrite(3+4, motorThrottle);
      ledcWrite(4+4, 0);
    }else{
      ledcWrite(3+4, 0);
      ledcWrite(4+4, -motorThrottle);    
    }
  }

  //write both motors in the same function call
  void writeMotors(int motorL, int motorR){
    writeMotorL(motorL);
    writeMotorR(motorR);
  }
}
